# Perform simple data binarization using the `infer_proteins` function
import numpy as np
from harissa import NetworkModel
from harissa.inference import infer_proteins
from harissa.utils import binarize

# Path of result files
result_path = "../examples/results/"

# Import raw data (run network4.py)
data = np.loadtxt(result_path + "network4_data.txt", dtype=int, delimiter='\t')
C, G = data.shape

# Store binarized data
newdata = np.zeros((C,G), dtype='int')
newdata[:,0] = data[:,0] # Time points

# Calibrate the mechanistic model
model = NetworkModel()
model.get_kinetics(data)

# Get binarized values (gene-specific thresholds)
newdata[:,1:] = infer_proteins(data, model.a)[:,1:].astype(int)

# Save binarized data
np.savetxt(result_path + "test_binarize.txt", newdata,
    fmt='%d', delimiter='\t')

# Note: a wrapper function is available
bdata = binarize(data)
print(f"newdata = binarize(data) → {np.array_equal(newdata, bdata)}")



"""
harissa.inference
-----------------

Inference of the network model.
"""
from harissa.inference.kinetics import infer_kinetics
from harissa.inference.network import infer_proteins

__all__ = ["infer_kinetics", "infer_proteins"]



"""
harissa.utils
-------------

Various utility functions.
"""
from harissa.utils.plot_network import build_pos, plot_network
from harissa.utils.processing import binarize

__all__ = ["build_pos", "plot_network", "binarize"]



# BurstyBase model (reduced/hybrid/bursty - single gene - no feedback)
import numpy as np
import matplotlib.pyplot as plt
from harissa.models import BurstyBase

# Path of result files
result_path = "../examples/results/bursty_base_"

model = BurstyBase()
model.burst_size = 0.5
model.burst_frequency = 2.5
model.degradation_rate = 0.8

# Compute a single trajectory
time = np.linspace(0, 10, 1000)
sim = model.simulate(time, init_state=1)

# Show the trajectory
fig = plt.figure(figsize=(12,3))
plt.plot(sim.t, sim.x)
plt.xlim(sim.t[0], sim.t[-1])
plt.ylim(0)
fig.savefig(result_path + "traj.pdf", bbox_inches='tight')

# Time-dependent distribution
x = np.linspace(0, 3, 1000)
init_state = 1
time = np.linspace(0, 1.6, 3)
p = model.distribution(x, time, init_state)

# Confirm analytical formula
t_simul = time[1]
n_cells = 10000
cell_pop = np.zeros(n_cells)
for k in range(n_cells):
    sim = model.simulate(t_simul, init_state)
    cell_pop[k] = sim.x[0]

# Show time-dependent distribution
fig = plt.figure(figsize=(8,3))
for k, t in enumerate(time):
    plt.plot(x, p[k], label=f"t = {t}")
    print(f"t = {t} -> non-zeros: {np.count_nonzero(p[k])}")
plt.hist(cell_pop, density=True, bins=100, color='lightgray', zorder=0,
    label=f"Simulation (t = {t_simul})")
plt.xlim(x[0], x[-1])
plt.ylim(0, 1)
plt.legend(loc='upper right')
fig.savefig(result_path + "dist.pdf", bbox_inches='tight')



"""
Various checks for function parameters within the API.
"""
import numpy as np

def _check_time_points(time):
    """
    Check time points for trajectory simulations.
    """
    if np.ndim(time) == 0:
        time = np.array([time], dtype=float)
    elif np.ndim(time) == 1:
        time = np.array(time, dtype=float)
    else:
        raise ValueError("Time points should either be scalar or 1D array.")
    if np.any(time != np.sort(time)):
        raise ValueError("Time points must be given in increasing order.")
    if time[0] < 0:
        raise ValueError("Time points must be nonnegative.")
    return time

def _check_init_state(init_state, shape=None):
    """
    Check initial state for trajectory simulations.
    """
    if isinstance(shape, tuple) and (np.shape(init_state) != shape):
        if shape == ():
            raise ValueError("Initial state must be a scalar value.")
        else:
            raise ValueError(f"Initial state must have shape {shape}.")
    if np.any(np.array(init_state) < 0):
        raise ValueError("Initial state must be nonnegative.")
    return init_state


# Tests
if __name__ == '__main__':
    time = np.linspace(0, 10, 100)
    _check_time_points(time)
    state = 0, 1
    _check_init_state(state)



"""
Data processing
"""
import numpy as np
from harissa.model import NetworkModel
from harissa.inference import infer_proteins

def binarize(data):
    """
    Return a binarized version of the data using gene-specific thresholds
    derived from the data-calibrated mechanistic model.
    """
    new_data = np.zeros(data.shape, dtype=data.dtype)
    new_data[:,0] = data[:,0] # Time points
    # Calibrate the mechanistic model
    model = NetworkModel()
    model.get_kinetics(data)
    # Get binarized values (gene-specific thresholds)
    new_data[:,1:] = infer_proteins(data, model.a)[:,1:].astype(int)
    return new_data



"""
Store simulations
"""

class Simulation:
    """
    Basic object to store simulations.
    """
    def __init__(self, t, m, p):
        self.t = t # Time points
        self.m = m # mRNAs
        self.p = p # Proteins



"""
Reduced/Hybrid/Bursty model for a single gene with no feedback.
"""
import numpy as np
from scipy.special import hyp1f1
from harissa.models._checks import _check_time_points, _check_init_state

class Simulation:
    """
    Basic object to store simulations.
    """
    def __init__(self, t, x):
        self.t = t # Time points
        self.x = x # Proteins or mRNA

class BurstyBase:
    """
    Reduced-Hybrid-Bursty model for a single gene with no feedback.
    NB: For this model the propagator is available in analytical form.
    """
    def __init__(self,
        burst_size=1.0,
        burst_frequency=1.0,
        degradation_rate=1.0):
        # Set model parameters
        self.burst_size = burst_size
        self.burst_frequency = burst_frequency
        self.degradation_rate = degradation_rate

    def simulate(self, time, init_state=0.0, verb=False):
        """
        Exact simulation of the model (extracted at given time points).
        """
        burst_size = self.burst_size
        burst_frequency = self.burst_frequency
        degradation_rate = self.degradation_rate
        # Check parameters
        time = _check_time_points(time)
        init_state = _check_init_state(init_state)
        # Number of bursts
        n = np.random.poisson(burst_frequency * time[-1])
        # Burst times and heights
        t = np.random.uniform(low=0, high=time[-1], size=n)
        h = np.random.exponential(scale=burst_size, size=n)
        # Build the post-jump embedded Markov chain
        t = np.append(0, np.sort(t))
        x = np.zeros(n + 1)
        x[0] = init_state
        for k in range(n):
            dt = t[k+1] - t[k]
            x[k+1] = x[k] * np.exp(- degradation_rate * dt) + h[k]
        # Extract states at user time points
        traj = np.zeros(time.size)
        for i, u in enumerate(time):
            k = np.argwhere(t <= u)[-1][0]
            dt = u - t[k]
            traj[i] = x[k] * np.exp(- degradation_rate * dt)
        if verb:
            print(f'Simulation generated {n} jumps.')
        return Simulation(time, traj)

    def distribution(self, x, time, init_state=0.0, smooth=1e-2):
        """
        Time-dependent distribution conditionally to initial_state.
        """
        b = 1 / self.burst_size
        k = self.burst_frequency
        d = self.degradation_rate
        # Check parameters
        t = _check_time_points(time).reshape((-1, 1))
        x0 = _check_init_state(init_state, shape=())
        # Smoothing standard deviation
        s = smooth
        # State when no burst occurred
        xt = np.exp(- d * t) * x0
        # Array of translated states
        x = x - xt
        # Scaling factors
        e0 = np.exp(d * t)
        e1 = np.exp(k * t)
        e1[t == 0] = 0 # Discarded
        r1 = b * (e0 - 1)
        r2 = (k/d) * (x >= 0) / (e1 - 1)
        # Probability that no burst occurred
        w = np.exp(- k * t)
        # Pre-burst part of the distribution (regularized)
        h = np.exp(- x**2 / (2 * s**2)) / (np.sqrt(2 * np.pi) * s)
        # Post-burst part of the distribution
        # # Option 1: gamma-mixture formula
        # g = r1 * r2 * np.exp(- b * e0 * x) * hyp1f1(k/d + 1, 2, r1 * x)
        # Option 2: using Kummer's transformation
        g = r1 * r2 * np.exp(- b * x) * hyp1f1(1 - k/d, 2, -r1 * x)
        # Overall distribution
        p = w * h + (1 - w) * g
        if t.size == 1:
            p = p[0]
        return p


# Tests
if __name__ == '__main__':
    model = BurstyBase()
    # Simulation
    time = np.linspace(0, 50, 6)
    sim = model.simulate(time, verb=True)
    print(sim.t)
    print(sim.x)
    # Distribution
    x = np.linspace(0, 2, 4)
    time = 0, 1, 2
    init_state = 0.5
    p = model.distribution(x, time, init_state)
    print(p)



# Basic 4-gene network with stimulus and feedback loop
import numpy as np
import matplotlib.pyplot as plt
from harissa import NetworkModel
from harissa.utils import build_pos, plot_network

# Path of result files
result_path = "../examples/results/network4_"


#### Simulate scRNA-seq data ####

# Total number of cells
C = 1000

# Set the time points
k = np.linspace(0, C, 11, dtype='int')
t = np.linspace(0, 20, 10, dtype='int')
time = np.zeros(C, dtype='int')
for i in range(10):
    time[k[i]:k[i+1]] = t[i]
print(f"Times points ({t.size}): {t}")

# Number of genes
G = 4

# Prepare data
data = np.zeros((C, G+1), dtype='int')
data[:,0] = time # Time points

# Initialize the model
model = NetworkModel(G)
model.d[0] = 1
model.d[1] = 0.2
model.basal[1:] = -5
model.inter[0,1] = 10
model.inter[1,2] = 10
model.inter[1,3] = 10
model.inter[3,4] = 10
model.inter[4,1] = -10
model.inter[2,2] = 10
model.inter[3,3] = 10

# Generate data
for k in range(C):
    print(f"* Cell {k+1} (t = {time[k]})")
    sim = model.simulate(time[k], burnin=5)
    data[k,1:] = np.random.poisson(sim.m[0])

# Save data in basic format
np.savetxt(result_path + "data.txt", data, fmt='%d', delimiter='\t')


#### Plot mean trajectory ####

# Import time points
time = np.sort(list(set(data[:,0])))
T = np.size(time)

# Average for each time point
traj = np.zeros((T,G))
for k, t in enumerate(time):
    traj[k] = np.mean(data[data[:,0]==t,1:], axis=0)

# Draw trajectory and export figure
fig = plt.figure(figsize=(8,2))
labels = [rf"$\langle M_{i+1} \rangle$" for i in range(G)]
plt.plot(time, traj, label=labels)
ax = plt.gca()
ax.set_xlim(time[0], time[-1])
ax.set_ylim(0, 1.2*np.max(traj))
ax.set_xticks(time)
ax.set_title(f"Bulk-average trajectory ({int(C/T)} cells per time point)")
ax.legend(loc='upper left', ncol=G, borderaxespad=0, frameon=False)
fig.savefig(result_path + "mean.pdf", bbox_inches='tight')


#### Plot the network ####

# Node labels and positions
names = [''] + [f'{i+1}' for i in range(G)]
pos = build_pos(model.inter)

# Draw network and export figure
fig = plt.figure(figsize=(5,5))
plot_network(model.inter, pos, axes=fig.gca(), names=names, scale=2)
fig.savefig(result_path + "graph.pdf", bbox_inches='tight')


#### Perform network inference ####

# Load the data
x = np.loadtxt(result_path + "data.txt", dtype=int, delimiter='\t')

# Calibrate the model
model = NetworkModel()
model.fit(x)

# Export interaction matrix
np.savetxt(result_path + "inter.txt", model.inter, delimiter='\t')



"""
Inference of basal parameters
"""
import numpy as np
from numpy import log
from scipy.special import psi, polygamma

def estim_gamma(x):
    """
    Estimate the parameters of a gamma distribution using
    the method of moments. The output is (a,b) for the distribution
    f(x) = x**(a-1)*exp(-b*x)/(gamma(a)/b**a).
    """
    m = np.mean(x)
    v = np.var(x)
    if v == 0:
        return 0, 1
    else:
        return m*m/v, m/v

def estim_gamma_poisson(x):
    """
    Estimate parameters a and b of the Gamma-Poisson(a,b) distribution,
    a.k.a. negative binomial distribution, using the method of moments.
    """
    m1 = np.mean(x)
    m2 = np.mean(x*(x-1))
    if m1 == 0:
        return 0, 1
    r = m2 - m1**2
    if r > 0:
        b = m1/r
    else:
        v = np.var(x)
        if v == 0:
            return 0, 1
        b = m1/v
    a = m1 * b
    return a, b

def transform(x):
    """
    Replace x by the conditional expectation given x of the underlying
    Gamma distribution, within the Gamma-Poisson model inferred from x.
    NB: this simply corresponds to a linear transformation with offset.
    """
    a, b = estim_gamma_poisson(x)
    if not (a > 0 and b > 0):
        print(("Warning: you should check whether x is not "
            f"almost zero (sum(x) = {np.sum(x)})."))
        a, b = np.abs(a), np.abs(b)
    return (a + x)/(b + 1)

def infer_kinetics(x, times, tol=1e-5, max_iter=100, verb=False):
    """
    Infer parameters a[0], ..., a[m-1] and b of a Gamma-Poisson model
    with time-dependant a and constant b for a given gene at m time points.

    Parameters
    ----------
    x : ndarray
        `x[k]` is the gene expression of cell `k`
    times : ndarray
        `times[k]` is the time point of cell `k`
    """
    t = np.sort(list(set(times)))
    m = t.size
    n = np.zeros(m) # Number of cells for each time point
    a = np.zeros(m)
    b = np.zeros(m)
    # Initialization of a and b
    for i in range(m):
        cells = (times == t[i])
        n[i] = np.sum(cells)
        a[i], b[i] = estim_gamma_poisson(x[cells])
    b = np.mean(b)
    # Newton-like method
    k, c = 0, 0
    sx = np.sum(x)
    while (k == 0) or (k < max_iter and c > tol):
        da = np.zeros(m)
        for i in range(m):
            if a[i] > 0:
                cells = (times == t[i])
                z = a[i] + x[cells]
                p0 = np.sum(psi(z))
                p1 = np.sum(polygamma(1, z))
                d = n[i]*(log(b)-log(b+1)-psi(a[i])) + p0
                h = p1 - n[i]*polygamma(1, a[i])
                da[i] = -d/h
        anew = a + da
        if np.sum(anew < 0) == 0:
            a[:] = anew
        else:
            max_test = 5
            test = 0
            da *= 0.5
            while (np.sum(a + da < 0) > 0) and (test < max_test):
                da *= 0.5
                test += 1
            if test < max_test:
                a[:] = a + da
            else:
                print("Warning: parameter a not improved")
        if np.sum(a == 0) == 0:
            b = np.sum(n*a)/sx
        else:
            b = 1
        c = np.max(np.abs(da))
        k += 1
    if (k == max_iter) and (c > tol):
        # print(f"Warning: bad convergence (b = {b})")
        a, b = a/b, 1
    if verb:
        print(f"Estimation done in {k} iterations")
    if np.sum(a < 0) > 0:
        print("WARNING: a < 0")
    if b < 0:
        print("WARNING: b < 0")
    if np.all(a == 0):
        print("WARNING: a == 0")
    # if k > 20 and np.max(a/b) > 2: print(k, np.max(a/b))
    return a, b


# Tests
if __name__ == '__main__':
    x = np.array([2,0,10,5,0,7])
    times = np.array([0,0,0,1,1,1])
    a, b = infer_kinetics(x, times, verb=True)
    print(a, b)



# Example of network layout using harissa.utils
import numpy as np
import matplotlib.pyplot as plt
from harissa.utils import build_pos, plot_network

# Path of result files
result_path = "../examples/results/"

# Interaction matrix
inter = np.array([
    [ 1, 1, 1],
    [-1,-1,-1],
    [ 1,-1, 1]])

# Node labels and positions
names = [f"$G_{{{i+1}}}$" for i in range(inter[0].size)]
# Option 1: layout from networkx
pos = build_pos(inter)
# Option 2: layout from networkx + graphviz (needs installation)
# pos = build_pos(inter, method='graphviz')

# Figure
fig = plt.figure(figsize=(5,5))
ax = fig.gca()

# Draw the network
plot_network(inter, pos, axes=ax, names=names, scale=2)

# Export the figure
fig.savefig(result_path + "test_plot_network.pdf", bbox_inches='tight')



"""
Perform simulations using the PDMP model
"""
import numpy as np
from scipy.special import expit

class BurstyPDMP:
    """
    Bursty PDMP version of the network model (promoters not described)
    """
    def __init__(self, a, d, basal, inter, thin_adapt=True):
        # Kinetic parameters
        G = basal.size
        D0, D1 = d[0], d[1]
        K0, K1, B = a[0]*d[0], a[1]*d[0], a[2]
        S1 = D0*D1*a[2]/K1 # Normalize protein scales
        types = [('D0', 'float'), ('D1', 'float'), ('S1', 'float'),
            ('K0', 'float'), ('K1', 'float'), ('B', 'float')]
        plist = [(D0[i], D1[i], S1[i], K0[i], K1[i], B[i]) for i in range(G)]
        self.param = np.array(plist, dtype=types)
        # Network parameters
        self.basal = basal
        self.inter = inter
        # Default state
        types = [('M', 'float'), ('P', 'float')]
        self.state = np.array([(0,0) for i in range(G)], dtype=types)
        # Thinning parameter
        self.thin_cst = None if thin_adapt else np.sum(K1[1:])

    def kon(self, p):
        """
        Interaction function kon (off->on rate), given protein levels p.
        """
        K0, K1 = self.param['K0'], self.param['K1']
        sigma = expit(self.basal + p @ self.inter)
        Kon = (1-sigma)*K0 + sigma*K1
        Kon[0] = 0 # Ignore stimulus
        return Kon

    def kon_bound(self):
        """
        Compute the current kon upper bound.
        """
        M, P = self.state['M'], self.state['P']
        D0, D1, S1 = self.param['D0'], self.param['D1'], self.param['S1']
        # Explicit upper bound for P
        time = np.log(D0/D1)/(D0-D1) # Vector of critical times
        pmax = P + (S1/(D0-D1))*M*(np.exp(-time*D1) - np.exp(-time*D0))
        pmax[0] = P[0] # Discard stimulus
        # Explicit upper bound for Kon
        K0, K1 = self.param['K0'], self.param['K1']
        sigma = expit(self.basal + pmax @ ((self.inter > 0) * self.inter))
        Kon = (1-sigma)*K0 + sigma*K1 + 1e-10 # Fix precision errors
        Kon[0] = 0 # Ignore stimulus
        return Kon

    def flow(self, time, state):
        """
        Deterministic flow for the bursty model.
        """
        M, P = state['M'], state['P']
        D0, D1, S1 = self.param['D0'], self.param['D1'], self.param['S1']
        # Explicit solution of the ODE generating the flow
        Mnew = M*np.exp(-time*D0)
        Pnew = ((S1/(D0-D1))*M*(np.exp(-time*D1) - np.exp(-time*D0))
                + P*np.exp(-time*D1))
        Mnew[0], Pnew[0] = M[0], P[0] # Discard stimulus
        return Mnew, Pnew

    def step(self):
        """
        Compute the next jump and the next step of the
        thinning method, in the case of the bursty model.
        """
        if self.thin_cst is None:
            # Adaptive thinning parameter
            tau = np.sum(self.kon_bound())
        else:
            tau = self.thin_cst

        # 0. Draw the waiting time before the next jump
        U = np.random.exponential(scale=1/tau)

        # 1. Update the continuous states
        M, P = self.flow(U, self.state)
        self.state['M'], self.state['P'] = M, P

        # 2. Compute the next jump
        v = self.kon(P)/tau # i = 1, ..., G-1 : burst of mRNA i
        v[0] = 1.0 - np.sum(v[1:]) # i = 0 : no change (phantom jump)
        # Deal robustly with precision errors
        i = np.searchsorted(np.cumsum(v), np.random.random(), side='right')
        jump = i > 0 # Test if jump is a true (i > 0) or phantom jump (i == 0)
        if jump:
            r = self.param['B'][i]
            self.state['M'][i] += np.random.exponential(1/r)

        return U, jump

    def simulation(self, timepoints, verb=False):
        """
        Exact simulation of the network in the bursty PDMP case.
        """
        G = self.basal.size
        types = [('M', 'float64'), ('P', 'float64')]
        sim = [] # List of states to be recorded
        c0, c1 = 0, 0 # Jump counts (phantom and true)
        T = 0
        # Core loop for simulation and recording
        Told, state_old = T, self.state.copy()
        for t in timepoints:
            while T < t:
                Told, state_old = T, self.state.copy()
                U, jump = self.step()
                T += U
                if jump:
                    c1 += 1
                else:
                    c0 += 1
            M, P = self.flow(t - Told, state_old)
            sim += [np.array([(M[i],P[i]) for i in range(1,G)], dtype=types)]
        # Update the current state
        self.state['M'], self.state['P'] = M, P
        # Display info about jumps
        if verb:
            ctot = c0 + c1
            if ctot > 0:
                print((f"Exact simulation used {ctot} jumps "
                    f"including {c0} phantom jumps ({100*c0/ctot:.2f}%)"))
            else:
                print("Exact simulation used no jump")
        return np.array(sim)



# Basic repressilator network (3 genes)
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from harissa import NetworkModel

# Path of result files
result_path = "../examples/results/"

# Model
model = NetworkModel(3)
model.d[0] = 1
model.d[1] = 0.2
model.basal[1] = 5
model.basal[2] = 5
model.basal[3] = 5
model.inter[1,2] = -10
model.inter[2,3] = -10
model.inter[3,1] = -10

# Time points
time = np.linspace(0,100,1000)

# Simulation of the PDMP model
sim = model.simulate(time)

# Simulation of the ODE model (slow-fast limit)
sim_ode = model.simulate_ode(time, P0=[0,0,0.1,0.2])

# Figure
fig = plt.figure(figsize=(10,6))
gs = gridspec.GridSpec(3, 1, hspace=0.6)

# Plot mRNA levels
ax = plt.subplot(gs[0,0])
for i in range(3):
    ax.set_title(f"mRNA levels ($d_0 = {model.d[0].mean()}$)")
    ax.plot(sim.t, sim.m[:,i], label=f"$M_{{{i+1}}}$")
    ax.set_xlim(sim.t[0], sim.t[-1])
    ax.set_ylim(0, 1.2*np.max(sim.m))
    ax.legend(loc='upper left', ncol=4, borderaxespad=0, frameon=False)

# Plot protein levels
ax = plt.subplot(gs[1,0])
for i in range(3):
    ax.set_title(f"Protein levels ($d_1 = {model.d[1].mean()}$)")
    ax.plot(sim.t, sim.p[:,i], label=f"$P_{{{i+1}}}$")
    ax.set_xlim(sim.t[0], sim.t[-1])
    ax.set_ylim(0, np.max([1.2*np.max(sim.p), 1]))
    ax.legend(loc='upper left', ncol=4, borderaxespad=0, frameon=False)

# Plot protein levels (ODE model)
ax = plt.subplot(gs[2,0])
for i in range(3):
    ax.set_title(r"Protein levels - ODE model ($d_0/d_1\to\infty$)")
    ax.plot(sim_ode.t, sim_ode.p[:,i], label=f"$P_{{{i+1}}}$")
    ax.set_xlim(sim_ode.t[0], sim_ode.t[-1])
    ax.set_ylim(0,1)
    ax.legend(loc='upper left', ncol=4, borderaxespad=0, frameon=False)

fig.savefig(result_path + "repressilator.pdf", bbox_inches='tight')



"""
Generate random trees
"""
import numpy as np

def random_step(state, a):
    """
    Make one step of the random walk on the weighted graph defined by a.
    NB: here we construct an in-tree so all directions are reversed.
    """
    p = a[:,state]/np.sum(a[:,state])
    return np.dot(np.arange(p.size), np.random.multinomial(1, p))

def loop_erasure(path):
    """
    Compute the loop erasure of a given path.
    """
    if path[0] == path[-1]:
        return [path[0]]
    else:
        i = np.max(np.arange(len(path))*(np.array(path)==path[0]))
    if path[i+1] == path[-1]:
        return [path[0], path[i+1]]
    else:
        return [path[0]] + loop_erasure(path[i+1:])

def random_tree(a):
    """
    Generate a random spanning tree rooted in node 0 from the uniform
    distribution with weights given by matrix a (using Wilson’s method).
    """
    n = a[0].size
    tree = [[] for i in range(n)]
    v = {0} # Vertices of the tree
    r = list(range(1,n)) # Remaining vertices
    while len(r) > 0:
        state = r[0]
        path = [state]
        # compute a random path that reaches the current tree
        while path[-1] not in v:
            state = random_step(path[-1], a)
            path.append(state)
        path = loop_erasure(path)
        # Append the loop-erased path to the current tree
        for i in range(len(path)-1):
            v.add(path[i])
            r.remove(path[i])
            tree[path[i+1]].append(path[i])
    for i in range(n):
        tree[i].sort()
    return tuple([tuple(tree[i]) for i in range(n)])

# Main function
def tree(n_genes, weight=None):
    """
    Generate a random tree-like network model.
    A tree with root 0 is sampled from the ‘weighted-uniform’ distribution,
    where weight[i,j] is the probability weight of link (i) -> (j).
    """
    G = n_genes + 1
    if weight is not None:
        if weight.shape != (G,G):
            raise ValueError(("weight must have shape "
                "(n_genes + 1, n_genes + 1)"))
    else:
        weight = np.ones((G,G))
    # Enforcing the proper structure
    weight[:,0] = 0
    weight = weight - np.diag(np.diag(weight))
    # Generate the network
    tree = random_tree(weight)
    basal = np.zeros(G)
    inter = np.zeros((G,G))
    basal[1:] = -5
    for i, targets in enumerate(tree):
        for j in targets:
            inter[i,j] = 10
    return basal, inter


# Tests
if __name__ == '__main__':
    basal, inter = tree(5)
    print(basal)
    print(inter)



"""
Generate cascade networks
"""
import numpy as np

def cascade(n_genes):
    """
    Generate a simple activation cascade (1) -> (2) -> ... -> (n_genes).
    """
    G = n_genes + 1
    basal = np.zeros(G)
    inter = np.zeros((G, G))
    basal[1:] = -5 # Low basal level of downstream genes
    for i in range(n_genes):
        inter[i, i+1] = 10
    return basal, inter


# Tests
if __name__ == '__main__':
    basal, inter = cascade(5)
    print(inter)



"""
Main class for network inference and simulation
"""
import numpy as np
from harissa.model.cascade import cascade
from harissa.model.tree import tree
from harissa.inference import infer_kinetics, infer_proteins

class NetworkModel:
    """
    Handle networks within Harissa.
    """
    def __init__(self, n_genes=None):
        # Kinetic parameters
        self.a = None
        self.d = None
        # Network parameters
        self.basal = None
        self.inter = None
        # Default behaviour
        if n_genes is not None:
            G = n_genes + 1 # Genes plus stimulus
            # Default bursting parameters
            self.a = np.zeros((3,G))
            self.a[0] = 0 # Minimal Kon rate (normalized)
            self.a[1] = 2 # Maximal Kon rate (normalized)
            self.a[2] = 0.02 # Inverse burst size of mRNA
            # Default degradation rates
            self.d = np.zeros((2,G))
            self.d[0] = np.log(2)/9 # mRNA degradation rates (per hour)
            self.d[1] = np.log(2)/46 # protein degradation rates (per hour)
            # Default network parameters
            self.basal = np.zeros(G)
            self.inter = np.zeros((G,G))

    def get_kinetics(self, data, verb=False):
        """
        Compute the basal parameters of all genes.
        """
        times = data[:,0]
        G = data[0].size
        # Kinetic values for each gene
        a = np.ones((3,G))
        for g in range(1,G):
            if verb:
                print(f"Calibrating gene {g}...")
            x = data[:,g]
            at, b = infer_kinetics(x, times, verb=verb)
            a[0,g] = np.min(at)
            a[1,g] = np.max(at)
            a[2,g] = b
        self.a = a

    def fit(self, data, l=1, tol=1e-5, verb=False, use_numba=True):
        """
        Fit the network model to the data.
        """
        x = data
        # Time points
        times = np.sort(list(set(x[:,0])))
        self.times = times
        # Default degradation rates
        C, G = x.shape
        d = np.zeros((2,G))
        d[0] = np.log(2)/9 # mRNA degradation rates
        d[1] = np.log(2)/46 # protein degradation rates
        self.d = d
        # Kinetic parameters
        self.get_kinetics(data, verb)
        a = self.a
        # Concentration parameter
        c = 100 * np.ones(G)
        # Get protein levels
        y = infer_proteins(x, a)
        self.y = y
        # Import necessary modules
        if use_numba:
            from harissa.inference.network_fast import infer_network
        else:
            from harissa.inference.network import infer_network
        # Inference procedure
        theta = infer_network(x, y, a, c, l, tol, verb)
        # Build the results
        self.basal = np.zeros(G)
        self.inter = np.zeros((G,G))
        self.basal_time = {time: np.zeros(G) for time in times}
        self.inter_time = {time: np.zeros((G,G)) for time in times}
        for t, time in enumerate(times):
            self.basal_time[time][:] = theta[t][:,0]
            self.inter_time[time][:,1:] = theta[t][:,1:]
        self.basal[:] = theta[-1][:,0]
        self.inter[:,1:] = theta[-1][:,1:]

    def simulate(self, t, M0=None, P0=None, burnin=None, verb=False,
        use_numba=False):
        """
        Perform simulation of the network model (bursty PDMP version).
        """
        # Check parameters
        check = ((self.a is None) + (self.d is None)
                + (self.basal is None) + (self.inter is None))
        # Prepare time points
        if check:
            raise ValueError("Model parameters not yet specified")
        if np.size(t) == 1:
            t = np.array([t])
        if np.any(t != np.sort(t)):
            raise ValueError("Time points must appear in increasing order")
        # Import necessary modules
        from harissa.simulation.base import Simulation
        if use_numba:
            from harissa.simulation.pdmp_fast import BurstyPDMP
        else:
            from harissa.simulation.pdmp import BurstyPDMP
        a = self.a
        d = self.d
        basal = self.basal
        inter = self.inter
        network = BurstyPDMP(a, d, basal, inter)
        # Burnin simulation without stimulus
        if M0 is not None:
            network.state['M'][1:] = M0[1:]
        if P0 is not None:
            network.state['P'][1:] = P0[1:]
        if burnin is not None:
            network.simulation([burnin], verb)
        # Activate the stimulus
        network.state['P'][0] = 1
        # Final simulation with stimulus
        sim = network.simulation(t, verb)
        m, p = sim['M'], sim['P']
        return Simulation(t, m, p)

    def simulate_ode(self, t, M0=None, P0=None, burnin=None, verb=False):
        """
        Perform simulation of the network model (ODE version).
        This is the slow-fast limit of the PDMP model, which is only
        relevant when promoters & mRNA are much faster than proteins.
        p: solution of a nonlinear ODE system involving proteins only
        m: mean mRNA levels given protein levels (quasi-steady state)
        """
        # Check parameters
        check = ((self.a is None) + (self.d is None)
                + (self.basal is None) + (self.inter is None))
        # Prepare time points
        if check:
            raise ValueError("Model parameters not specified yet")
        if self.inter is None:
            print("Interactions must be specified")
        if np.size(t) == 1:
            t = np.array([t])
        if np.any(t != np.sort(t)):
            raise ValueError("Time points must appear in increasing order")
        # Import necessary modules
        from harissa.simulation.base import Simulation
        from harissa.simulation.ode import ApproxODE
        a = self.a
        d = self.d
        basal = self.basal
        inter = self.inter
        network = ApproxODE(a, d, basal, inter)
        # Burnin simulation without stimulus
        if M0 is not None:
            network.state['M'][1:] = M0[1:]
        if P0 is not None:
            network.state['P'][1:] = P0[1:]
        if burnin is not None:
            network.simulation([burnin], verb)
        # Activate the stimulus
        network.state['P'][0] = 1
        # Final simulation with stimulus
        sim = network.simulation(t, verb)
        m, p = sim['M'], sim['P']
        return Simulation(t, m, p)

#### Classes for simulations ####

class Cascade(NetworkModel):
    """
    Particular network with a cascade structure.
    """
    def __init__(self, n_genes, autoactiv=False):
        # Get NetworkModel default features
        NetworkModel.__init__(self, n_genes)
        # New network parameters
        basal, inter = cascade(n_genes)
        if autoactiv:
            for i in range(1, n_genes + 1):
                inter[i, i] = 5
        self.basal = basal
        self.inter = inter

class Tree(NetworkModel):
    """
    Random network with a tree structure.
    """
    def __init__(self, n_genes, autoactiv=False):
        # Get NetworkModel default features
        NetworkModel.__init__(self, n_genes)
        # New network parameters
        basal, inter = tree(n_genes)
        if autoactiv:
            for i in range(1, n_genes + 1):
                inter[i, i] = 5
        self.basal = basal
        self.inter = inter



"""
Core functions for network inference using likelihood maximization
NB: This is the fast version using Numba
"""
import numpy as np
from numpy import log
from scipy.special import expit, gammaln, psi
from scipy.optimize import minimize
from numba import njit

# Smoothing threshold
s = 0.1

@njit
def p1(x, s):
    """
    Smoothed L1 penalization.
    """
    return (x-s/2)*(x>s) - (x+s/2)*(-x>s) + ((x**2)/(2*s))*(x<=s and -x<=s)

@njit
def grad_p1(x, s):
    """
    Smoothed L1 penalization gradient.
    """
    return 1*(x>s) - 1*(-x>s) + (x/s)*(x<=s and -x<=s)

@njit
def penalization(theta, theta0, t):
    """
    Penalization of network parameters.
    """
    G = theta.shape[0]
    p = 0
    for i in range (1,G):
        # Penalization of basal parameters
        p += 2 * t * p1(theta[i,0]-theta0[i,0], s)
        # Penalization of stimulus parameters
        p += t * p1(theta[0,i]-theta0[0,i], s)
        # Penalization of diagonal parameters
        p += (theta[i,i]-theta0[i,i])**2
        for j in range(1,G):
            # Penalization of interaction parameters
            p += p1(theta[i,j]-theta0[i,j], s)
            if i < j:
                # Competition between interaction parameters
                p += p1(theta[i,j], s) * p1(theta[j,i], s)
    # Final penalization
    return p

@njit
def grad_penalization(theta, theta0, t):
    """
    Penalization gradient of network parameters.
    """
    G = theta.shape[0]
    gradp = np.zeros((G,G))
    for i in range (1,G):
        # Penalization of basal parameters
        gradp[i,0] += 2 * t * grad_p1(theta[i,0]-theta0[i,0], s)
        # Penalization of stimulus parameters
        gradp[0,i] += t * grad_p1(theta[0,i]-theta0[0,i], s)
        # Penalization of diagonal parameters
        gradp[i,i] += 2*(theta[i,i]-theta0[i,i])
        for j in range(1,G):
            # Penalization of interaction parameters
            gradp[i,j] += grad_p1(theta[i,j]-theta0[i,j], s)
            if i != j:
                # Competition between interaction parameters
                gradp[i,j] += grad_p1(theta[i,j], s) * p1(theta[j,i], s)
    # Final penalization
    return gradp

def objective(theta, theta0, x, y, a, c, d, l, t):
    """
    Objective function to be minimized (one time point).
    """
    C, G = x.shape
    theta = theta.reshape((G,G))
    basal = theta[:,0]
    sigma = expit(basal + y @ theta)[:,1:]
    x, y = x[:,1:], y[:,1:]
    a0, a1, c, d = a[0,1:], a[1,1:], c[1:], d[1:]
    # Compute the log-likelihood
    ay, e = a1*y, a0/a1
    cxi = c * (e + (1-e)*sigma)
    q = (d*ay + gammaln(ay + x) - gammaln(ay) - c*y
        + (cxi-1)*log(y) + log(c)*cxi - gammaln(cxi))
    return l*penalization(theta, theta0, t) - np.sum(q)/C

def grad_theta(theta, theta0, x, y, a, c, d, l, t):
    """
    Objective gradient (one time point).
    """
    C, G = x.shape
    theta = theta.reshape((G,G))
    basal = theta[:,0]
    sigma = expit(basal + y @ theta)[:,1:]
    a0, a1, c, d = a[0,1:], a[1,1:], c[1:], d[1:]
    # Pivotal vector u
    e = a0/a1
    xi = e + (1-e)*sigma
    u = c * sigma * (1-xi) * (log(c*y[:,1:]) - psi(c*xi))
    # Compute the objective gradient
    dq = np.zeros((G,G))
    # Basal parameters
    dq[1:,0] += np.sum(u, axis=0)
    # Interaction parameters
    dq[:,1:] += y.T @ u
    dq = l*grad_penalization(theta, theta0, t) - dq/C
    return dq.reshape(G**2)

def infer_proteins(x, a):
    """
    Estimate y directly from data.
    """
    C, G = x.shape
    y = np.ones((C,G))
    z = np.ones((2,G))
    z[0] = a[0]/a[1]
    z[z<1e-5] = 1e-5
    az = a[1]*z
    for k in range(C):
        v = az*log(a[2]/(a[2]+1)) + gammaln(az+x[k]) - gammaln(az)
        for i in range(1,G):
            y[k,i] = z[np.argmax(v[:,i]),i]
    # Stimulus off at t <= 0
    y[x[:,0]<=0,0] = 0
    return y

def infer_network(x, y, a, c, l, tol, verb):
    """
    Network inference procedure.
    """
    C, G = x.shape
    times = np.sort(list(set(x[:,0])))
    T = times.size
    # Useful quantities
    k = x[:,0]
    d = log(a[2]/(a[2]+1))
    # Initialization
    theta = np.zeros((T,G,G))
    theta0 = np.zeros((G,G))
    # Optimization parameters
    params = {'method': 'L-BFGS-B'}
    if tol is not None:
        params['tol'] = tol
    # Inference routine
    for t, time in enumerate(times):
        res = minimize(objective, theta0.reshape(G**2),
                args=(theta0, x[k==time], y[k==time], a, c, d, l, t),
                jac=grad_theta, **params)
        if not res.success:
            print(f"Warning: maximization failed (time {t})")
        # Update theta0
        theta0 = res.x.reshape((G,G))
        # Store theta at time t
        theta[t] = theta0
    if verb:
        print(f"Fitted theta in {res.nit} iterations")
    return theta



# Branching 4-gene pathways with stimulus
import numpy as np
import matplotlib.pyplot as plt
from harissa import NetworkModel
from harissa.utils import build_pos, plot_network

# Path of result files
result_path = "../examples/results/pathways_"


#### Simulate scRNA-seq data ####

# Total number of cells
C = 1000

# Set the time points
k = np.linspace(0, C, 11, dtype='int')
t = np.linspace(0, 20, 10, dtype='int')
time = np.zeros(C, dtype='int')
for i in range(10):
    time[k[i]:k[i+1]] = t[i]
print(f"Times points ({t.size}): {t}")

# Number of genes
G = 4

# Prepare data
data1 = np.zeros((C, G+1), dtype='int')
data1[:,0] = time # Time points
data2 = data1.copy()

# Model 1
model1 = NetworkModel(G)
model1.d[0] = 1
model1.d[1] = 0.2
model1.basal[1:] = -5
model1.inter[0,1] = 10
model1.inter[1,2] = 10
model1.inter[1,3] = 10
model1.inter[2,4] = 10

# Model 2
model2 = NetworkModel(G)
model2.d[0] = 1
model2.d[1] = 0.2
model2.basal[1:] = -5
model2.inter[0,1] = 10
model2.inter[1,2] = 10
model2.inter[1,3] = 10
model2.inter[3,4] = 10

# Generate data
for k in range(C):
    # Data for model 1
    sim1 = model1.simulate(time[k], burnin=5)
    data1[k,1:] = np.random.poisson(sim1.m[0])
    # Data for model 2
    sim2 = model2.simulate(time[k], burnin=5)
    data2[k,1:] = np.random.poisson(sim2.m[0])

# Save data in basic format
np.savetxt(result_path + "data1.txt", data1, fmt='%d', delimiter='\t')
np.savetxt(result_path + "data2.txt", data2, fmt='%d', delimiter='\t')


#### Plot mean trajectories ####

for i, data in [(1,data1),(2,data2)]:
    # Import time points
    time = np.sort(list(set(data[:,0])))
    T = np.size(time)
    # Average for each time point
    traj = np.zeros((T,G))
    for k, t in enumerate(time):
        traj[k] = np.mean(data[data[:,0]==t,1:], axis=0)
    # Draw trajectory and export figure
    fig = plt.figure(figsize=(8,2))
    labels = [rf"$\langle M_{i+1} \rangle$" for i in range(G)]
    plt.plot(time, traj, label=labels)
    ax = plt.gca()
    ax.set_xlim(time[0], time[-1])
    ax.set_ylim(0, 1.2*np.max(traj))
    ax.set_xticks(time)
    ax.set_title(f"Bulk-average trajectory ({int(C/T)} cells per time point)")
    ax.legend(loc='upper left', ncol=G, borderaxespad=0, frameon=False)
    fig.savefig(result_path + f"mean{i}.pdf", bbox_inches='tight')


#### Plot the networks ####

# Node labels
names = [""] + [f"{i+1}" for i in range(G)]

# Draw networks and export figures
for i, model in [(1,model1),(2,model2)]:
    pos = build_pos(model.inter)
    fig = plt.figure(figsize=(5,5))
    plot_network(model.inter, pos, axes=fig.gca(), names=names, scale=2)
    fig.savefig(result_path + f"graph{i}.pdf", bbox_inches='tight')


#### Perform network inference ####

for i in [1,2]:
    # Load the data
    x = np.loadtxt(result_path + f"data{i}.txt", dtype=int, delimiter='\t')
    # Calibrate the model
    model = NetworkModel()
    model.fit(x)
    # Export interaction matrix
    np.savetxt(result_path + f"inter{i}.txt", model.inter, delimiter='\t')



"""
Perform simulations using the ODE model
"""
import numpy as np
from scipy.special import expit

class ApproxODE:
    """
    ODE version of the network model (very rough approximation of the PDMP)
    """
    def __init__(self, a, d, basal, inter):
        # Kinetic parameters
        G = basal.size
        D0, D1 = d[0], d[1]
        K0, K1, B = a[0]*d[0], a[1]*d[0], a[2]
        S1 = D0*D1*a[2]/K1 # Normalize protein scales
        types = [('D0', 'float'), ('D1', 'float'), ('S1', 'float'),
            ('K0', 'float'), ('K1', 'float'), ('B', 'float')]
        plist = [(D0[i], D1[i], S1[i], K0[i], K1[i], B[i]) for i in range(G)]
        self.param = np.array(plist, dtype=types)
        # Network parameters
        self.basal = basal
        self.inter = inter
        # Default state
        types = [('M', 'float'), ('P', 'float')]
        self.state = np.array([(0,0) for i in range(G)], dtype=types)
        # Simulation parameter
        self.euler_step = 1e-3/np.max(D1)

    def kon(self, p):
        """
        Interaction function kon (off->on rate), given protein levels p.
        """
        K0, K1 = self.param['K0'], self.param['K1']
        sigma = expit(self.basal + p @ self.inter)
        Kon = (1-sigma)*K0 + sigma*K1
        Kon[0] = 0 # Ignore stimulus
        return Kon

    def step_ode(self, dt):
        """
        Euler step for the deterministic limit model.
        """
        M, P = self.state['M'], self.state['P']
        D0, D1, S1 = self.param['D0'], self.param['D1'], self.param['S1']
        a, b = self.kon(P)/D0, self.param['B'] # a = kon/d0, b = koff/s0
        Mnew = a/b # Mean level of mRNA given protein levels
        Pnew = (1 - dt*D1)*P + dt*S1*Mnew # Protein-only ODE system
        Mnew[0], Pnew[0] = M[0], P[0] # Discard stimulus
        self.state['M'], self.state['P'] = Mnew, Pnew

    def simulation(self, timepoints, verb=False):
        """
        Simulation of the deterministic limit model, which is relevant when
        promoters and mRNA are much faster than proteins.
        1. Nonlinear ODE system involving proteins only
        2. Mean level of mRNA given protein levels
        """
        G = self.basal.size
        dt = self.euler_step
        if np.size(timepoints) > 1:
            dt = np.min([dt, np.min(timepoints[1:] - timepoints[:-1])])
        types = [('M', 'float64'), ('P', 'float64')]
        sim = []
        T, c = 0, 0
        # Core loop for simulation and recording
        for t in timepoints:
            while T < t:
                self.step_ode(dt)
                T += dt
                c += 1
            M, P = self.state['M'], self.state['P']
            sim += [np.array([(M[i],P[i]) for i in range(1,G)], dtype=types)]
        # Display info about steps
        if verb:
            if c > 0:
                print(f"ODE simulation used {c} steps "
                    + f"(step size = {dt:.5f})")
            else:
                print("ODE simulation used no step")
        return np.array(sim)
